/*  $Id: Cache.java 1188 2011-08-28 11:43:31Z sdevaux $
 *
 * Tabs = 4 chars
 * copyright (C) 2010 Sebastien Devaux
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 *              Sebastien Devaux <sebastien.devaux@laposte.net>
 */
package net.eduvax.util;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Random;
import java.util.Vector;

/**
 * Simple object cache.
 * This is a very simple cache : it does not deal with
 * object in memory size but only store a fixed count of
 * object, no matters with the size of objects.
 * TODO : make it thread safe...
 * @version $Id: Cache.java 1188 2011-08-28 11:43:31Z sdevaux $
 * @author Sebastien Devaux
 */ 
public class Cache<K,V> {
	/**
	 * Create new cache usign defaults LRU strategy.
	 * @param size cache size (max object count in cache).
	 * @param data data source
	 */
	public Cache(int size,DataSource<K,V> data) {
		init(size,data,new LRU<K>());
	}
	
	/**
	 * Create new cache.
	 * @param size cache size (max object count in cache).
	 * @param data data source
	 * @param strategy Cache mangement strategy.
	 */ 
	public Cache(int size,DataSource<K,V> data,CacheStrategy<K> cs) {
		init(size,data,cs);
	}
	
	/**
     * retrieve one object in cache.
     * return object in cache or get it from data source if not
     * yet cached.
	 * @return the object matching the key, null if not any.
	 */ 
	public V get(K key) {
		V res=_cachedObjs.get(key);
		if (res==null) {
			res=_dataSource.get(key);
			if (res!=null) {
				insert(key,res);
			}
		}
		else {
			_strategy.used(key);
		}
		return res;
	}
	/**
	 * insert object in cache.
	 * If cache is full, use strategy to determinate the object
	 * to remove from the cache. 
	 */ 
	private void insert(K key,V obj) {
		if (_cSize<_maxSize) {
			_cSize++;
		}
		else {
			K rmKey=_strategy.select();
			if (rmKey!=null) {
				_cachedObjs.remove(rmKey);
				_strategy.removed(rmKey);
			}
			else {
				App.get().log().warn("Cache strategy has not selected object to remove from cache.");
			}
		}
		_cachedObjs.put(key,obj);
		_strategy.added(key);
	}
	/** 
	 * Cache initialisation
	 * @param size cache size (max object count in cache).
	 * @param data data source
	 * @param strategy Cache mangement strategy.
	 */ 
	private void init(int size,DataSource<K,V> data,CacheStrategy<K> cs) {
		_maxSize=size;
		_cSize=0;
		_dataSource=data;
		_strategy=cs;
		_cachedObjs=new Hashtable<K,V>();
	}
	/** 
	 * Cached objects
	 * @djin.relation aggregation - * java.lang.Object
	 */
	private Hashtable<K,V> _cachedObjs;	
	/**
	 * Cache management sstrategy.
	 * @djin.relation composition - - net.eduvax.util.CacheStrategy
	 */
	private CacheStrategy<K> _strategy;	
	/** cache current size */
	private int _cSize;
	/** cache max size */
	private int _maxSize;
	/** data source */
	private DataSource<K,V> _dataSource;
	// -------------------------------------------------------------
	/**
	 * Default and trivial LRU implementation.
	 * Quick and Dirty LRU.
	 */
	static public class LRU<K> implements CacheStrategy<K> {
		/**
		 * Create new LRU
		 */
		public LRU() {
			_timestamps=new Hashtable<K,Long>();
			_useCount=0;
		}
		// ---- CacheStrategy interface implementation ----
		public K select() {
			long minUse=-1;
			K res=null;
			Enumeration<K> keys=_timestamps.keys();
			while (keys.hasMoreElements()) {
				K candidate=keys.nextElement();
				long use=_timestamps.get(candidate).longValue();
				if (minUse>use||minUse<0) {
					res=candidate;
					minUse=use;
				}
			}
			return res;
		}
		public void removed(K key) {
			_timestamps.remove(key);
		}
		public void used(K key) {
			_useCount++;
			_timestamps.remove(key);
			_timestamps.put(key,_useCount);
		}
		public void added(K key) {
			_useCount++;
			_timestamps.put(key,_useCount);
		}
		// ------------------------------------------------
		/**
		 * last date of use for each key.
		 * This does not really store dates but cache access counter snapshots.
		 * @djin.relation composition - * java.lang.Long
		 */ 
		private Hashtable<K,Long> _timestamps;
		/** Object uses counter */
		private long _useCount;
	}
	/**
	 * Trivial LFU implementation.
	 * Quick and Dirty LFU.
	 */
	static public class LFU<K> implements CacheStrategy<K> {
		/**
		 * Create new LRU
		 */
		public LFU() {
			_freqs=new Hashtable<K,Long>();
		}
		// ---- CacheStrategy interface implementation ----
		public K select() {
			long minUse=-1;
			K res=null;
			Enumeration<K> keys=_freqs.keys();
			while (keys.hasMoreElements()) {
				K candidate=keys.nextElement();
				long use=_freqs.get(candidate).longValue();
				if (minUse>use||minUse<0) {
					res=candidate;
					minUse=use;
				}
			}
			return res;
		}
		public void removed(K key) {
			_freqs.remove(key);
		}
		public void used(K key) {
			long freq=_freqs.remove(key).longValue();
			_freqs.put(key,freq+1);
		}
		public void added(K key) {
			_freqs.put(key,0l);
		}
		// ------------------------------------------------
		/**
		 * use  count for each key.
		 * @djin.relation composition - * java.lang.Long
		 */ 
		private Hashtable<K,Long> _freqs;
	}
	/**
	 * Trivial Random cache strategy.
	 * Quick and Dirty LFU.
	 */
	static public class Rand<K> implements CacheStrategy<K> {
		/**
		 * Create new LRU
		 */
		public Rand() {
			_rand=new Random();
			_cachedKeys=new Vector<K>();
		}
		// ---- CacheStrategy interface implementation ----
		public K select() {
			return _cachedKeys.elementAt(_rand.nextInt(_cachedKeys.size()));
		}
		public void removed(K key) {
			_cachedKeys.remove(key);
		}
		public void used(K key) {
		}
		public void added(K key) {
			_cachedKeys.add(key);
		}
		// ------------------------------------------------
		/**
		 * use  count for each key.
		 * @djin.relation composition - * java.lang.Long
		 */ 
		private Vector<K> _cachedKeys;
		/**
		 * Random generator
		 */
	   	private Random _rand;	
	}
}

